"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
}
Object.defineProperty(exports, "__esModule", { value: true });
var stable_1 = require("stable");
var query_string_1 = __importDefault(require("query-string"));
var HelpConfig;
(function (HelpConfig) {
    HelpConfig.minMaxWidth = 30;
})(HelpConfig = exports.HelpConfig || (exports.HelpConfig = {}));
function isShortName(name) {
    if (name.length <= 1) {
        return true;
    }
    if (name.length != 2) {
        return false;
    }
    var c = name[0];
    return c === '-' || c === '/';
}
var Formatter = /** @class */ (function () {
    function Formatter(config) {
        this._config = config;
        this._config.maxWidth = Math.max(config.maxWidth, HelpConfig.minMaxWidth);
        this._emittedChars = [];
        this._word = [];
        this._width = 0;
        this._indentation = 0;
    }
    Formatter.prototype.formatOption = function (option) {
        if (this._allowHelpGroup(option.helpGroup)) {
            this._width = 0;
            this._changeIndentation(0);
            this._newLine();
            this._formatOptionNames(option);
            this._formatOptionArgsText(option);
            this._formatOptionText(option);
            this._flushWord();
        }
    };
    Formatter.prototype.toString = function () {
        return this._emittedChars.join("");
    };
    Formatter.prototype._allowHelpGroup = function (helpGroup) {
        if (this._config.helpGroupFilter === undefined || this._config.helpGroupFilter.length === 0) {
            return true;
        }
        var idx = this._config.helpGroupFilter.indexOf(helpGroup);
        return idx >= 0;
    };
    Formatter.prototype._formatOptionNames = function (option) {
        var names = option.names;
        names.sort(function (n1, n2) {
            if (n1.length < n2.length) {
                return -1;
            }
            if (n1.length > n2.length) {
                return 1;
            }
            return 0;
        });
        if (names.length > 0) {
            var idx = 0;
            if (isShortName(names[idx])) {
                this._changeIndentation(1);
                this._emitString(names[idx]);
                ++idx;
            }
            for (; idx < names.length; ++idx) {
                if (idx > 0) {
                    this._emitChar(",");
                }
                this._changeIndentation(5);
                this._emitString(names[idx]);
            }
        }
    };
    Formatter.prototype._formatOptionArgsText = function (option) {
        var argsText = option.argsText();
        if (argsText.length > 0) {
            this._flushWord();
            this._changeIndentation(this._width + 1);
            this._emitString(argsText);
        }
    };
    Formatter.prototype._formatOptionText = function (option) {
        this._flushWord();
        if (option.text.length > 0) {
            this._changeIndentation(this._width + 1);
            this._changeIndentation(HelpConfig.minMaxWidth - 1);
            this._emitString(option.text);
        }
    };
    Formatter.prototype._flushWord = function () {
        if (this._word.length === 0) {
            return false;
        }
        if (!(this._width === this._indentation || this._word.length + this._width <= this._config.maxWidth)) {
            this._newLine(false);
        }
        for (var _i = 0, _a = this._word; _i < _a.length; _i++) {
            var c = _a[_i];
            this._emittedChars.push(c);
        }
        this._width += this._word.length;
        this._word.length = 0;
        return true;
    };
    Formatter.prototype._changeIndentation = function (newAmount) {
        this._flushWord();
        this._indentation = newAmount;
        this._indent();
    };
    Formatter.prototype._indent = function (flushWord) {
        if (flushWord === void 0) { flushWord = true; }
        if (flushWord) {
            this._flushWord();
        }
        if (this._width > this._indentation) {
            this._newLine();
            return;
        }
        var amount = this._indentation - this._width;
        for (var i = 0; i < amount; ++i) {
            this._emittedChars.push(" ");
        }
        this._width = this._indentation;
    };
    Formatter.prototype._newLine = function (flushWord) {
        if (flushWord === void 0) { flushWord = true; }
        if (this._emittedChars.length > 0) {
            this._emittedChars.push("\n");
        }
        this._width = 0;
        this._indent(flushWord);
    };
    Formatter.prototype._emitSpace = function () {
        if (this._flushWord()) {
            if (this._width < this._config.maxWidth) {
                this._emittedChars.push(" ");
                ++this._width;
            }
            else {
                this._newLine();
            }
        }
    };
    Formatter.prototype._emitChar = function (c) {
        if (c === " ") {
            this._emitSpace();
        }
        else {
            console.assert(c.length === 1);
            this._word.push(c);
        }
    };
    Formatter.prototype._emitString = function (str) {
        for (var _i = 0, str_1 = str; _i < str_1.length; _i++) {
            var c = str_1[_i];
            this._emitChar(c);
        }
    };
    return Formatter;
}());
exports.NoParse = Object.create(null);
function simpleParse(parser) {
    return function (startIndex, args) {
        if (args.length === startIndex) {
            return [exports.NoParse, 0];
        }
        var result = parser(args[startIndex]);
        if (result === exports.NoParse) {
            return [exports.NoParse, 0];
        }
        return [result, 1];
    };
}
exports.simpleParse = simpleParse;
function parseString(arg) {
    return arg;
}
exports.parseString = parseString;
function parseBoolean(arg) {
    if (arg === "true") {
        return true;
    }
    if (arg === "false") {
        return false;
    }
    return exports.NoParse;
}
exports.parseBoolean = parseBoolean;
function parseNumber(arg) {
    var n = Number(arg);
    if (isNaN(n)) {
        return exports.NoParse;
    }
    return n;
}
exports.parseNumber = parseNumber;
var isIntegerRegex = /^-?\d+$/;
var isUnsignedRegex = /^\d+$/;
function parseInteger(arg) {
    if (isIntegerRegex.test(arg)) {
        return parseInt(arg, 10);
    }
    return exports.NoParse;
}
exports.parseInteger = parseInteger;
function parseUnsigned(arg) {
    if (isUnsignedRegex.test(arg)) {
        return parseInt(arg, 10);
    }
    return exports.NoParse;
}
exports.parseUnsigned = parseUnsigned;
function createParseInt(bitCount) {
    var max = 1 << (bitCount - 1);
    var min = -max;
    return function (arg) {
        var n = parseInteger(arg);
        if (n === exports.NoParse || n >= max || n < min) {
            return exports.NoParse;
        }
        return n;
    };
}
function createParseUint(bitCount) {
    var max = 1 << bitCount;
    return function (arg) {
        var n = parseUnsigned(arg);
        if (n === exports.NoParse || n >= max) {
            return exports.NoParse;
        }
        return n;
    };
}
exports.parseInt8 = createParseInt(8);
exports.parseInt16 = createParseInt(16);
exports.parseInt32 = createParseInt(32);
exports.parseUint8 = createParseUint(8);
exports.parseUint16 = createParseUint(16);
exports.parseUint32 = createParseUint(32);
function createOptionalParser(parser) {
    return function (startIndex, args) {
        var result = parser(startIndex, args);
        if (result[0] === exports.NoParse) {
            result[0] = undefined;
            result[1] = 0;
        }
        return result;
    };
}
exports.createOptionalParser = createOptionalParser;
function createFixedArrayParser(sizeLiteral, parser) {
    var size = Number(sizeLiteral);
    return function (startIndex, args) {
        var parsedItems = [];
        var currIndex = startIndex;
        while (currIndex < args.length && parsedItems.length < size) {
            var _a = parser(currIndex, args), parsedItem = _a[0], consumedCount = _a[1];
            currIndex += consumedCount;
            if (parsedItem === exports.NoParse) {
                break;
            }
            parsedItems.push(parsedItem);
        }
        var totalConsumedCount = currIndex - startIndex;
        if (parsedItems.length === size) {
            return [parsedItems, totalConsumedCount];
        }
        return [exports.NoParse, totalConsumedCount];
    };
}
exports.createFixedArrayParser = createFixedArrayParser;
function createArrayParser(parser) {
    return function (startIndex, args) {
        var parsedItems = [];
        var currIndex = startIndex;
        var totalConsumedCount;
        while (currIndex < args.length) {
            var _a = parser(currIndex, args), parsedItem = _a[0], consumedCount = _a[1];
            if (parsedItem === exports.NoParse) {
                break;
            }
            parsedItems.push(parsedItem);
            currIndex += consumedCount;
        }
        totalConsumedCount = currIndex - startIndex;
        return [parsedItems, totalConsumedCount];
    };
}
exports.createArrayParser = createArrayParser;
function createUnionParser(parserA, parserB) {
    return function (startIndex, args) {
        var resultA = parserA(startIndex, args);
        if (resultA[0] !== exports.NoParse) {
            return resultA;
        }
        var resultB = parserB(startIndex, args);
        resultB[1] = Math.max(resultA[1], resultB[1]);
        return resultB;
    };
}
exports.createUnionParser = createUnionParser;
function createParsePackedArray(constructor, simpleParser) {
    var parser = createArrayParser(simpleParse(simpleParser));
    return function (startIndex, args) {
        var result = parser(startIndex, args);
        if (result[0] === exports.NoParse) {
            return result;
        }
        result[0] = new constructor(result[0]);
        return result;
    };
}
exports.parseInt8Array = createParsePackedArray(Int8Array, exports.parseInt8);
exports.parseInt16Array = createParsePackedArray(Int16Array, exports.parseInt16);
exports.parseInt32Array = createParsePackedArray(Int32Array, exports.parseInt32);
exports.parseUint8Array = createParsePackedArray(Uint8Array, exports.parseUint8);
exports.parseUint16Array = createParsePackedArray(Uint16Array, exports.parseUint16);
exports.parseUint32Array = createParsePackedArray(Uint32Array, exports.parseUint32);
//const parseFloat32Array = createParsePackedArray<Float32Array>(Float32Array, parseNumber);
//const parseFloat64Array = createParsePackedArray<Float64Array>(Float64Array, parseNumber);
var ArgInfo = /** @class */ (function () {
    function ArgInfo(typeName, argText) {
        this.typeName = typeName;
        this.argText = argText;
    }
    return ArgInfo;
}());
var Option = /** @class */ (function () {
    function Option(keywords, argInfos, text, helpGroup, func) {
        this.names = keywords;
        this.argInfos = argInfos;
        this.text = text;
        this.helpGroup = helpGroup;
        this.func = func;
    }
    Option.prototype.argsText = function () {
        return this.argInfos.map(function (info) {
            return info.argText;
        }).join(" ");
    };
    return Option;
}());
function isNameList(keyword) {
    return typeof keyword !== "string"
        && (keyword.length === 0
            || typeof keyword[0] === "string");
}
var OptionsImpl = /** @class */ (function () {
    function OptionsImpl() {
        this._parserMap = new Map();
        this._helpGroupPriorities = new Map();
        this._optionsByArity = new Map();
        this._maxArity = 0;
    }
    OptionsImpl.createRaw = function () {
        return new OptionsImpl();
    };
    OptionsImpl.prototype.registerParser = function (typeName, parser) {
        if (this._parserMap.has(typeName)) {
            throw new RegistrationException(typeName);
        }
        this._parserMap.set(typeName, parser);
        return this;
    };
    OptionsImpl.prototype.getParser = function (typeName) {
        return this._parserMap.get(typeName);
    };
    OptionsImpl.prototype.addOption = function (keyword) {
        var opaqueArgs = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            opaqueArgs[_i - 1] = arguments[_i];
        }
        var func = opaqueArgs.pop();
        var text = opaqueArgs.pop();
        var argInfos = [];
        for (var _a = 0, _b = opaqueArgs; _a < _b.length; _a++) {
            var opaqueArg = _b[_a];
            var typeName = void 0;
            var argText = void 0;
            if (opaqueArg instanceof Array) {
                typeName = opaqueArg[0];
                argText = opaqueArg[1];
            }
            else {
                typeName = opaqueArg.typeName;
                argText = opaqueArg.argText;
            }
            var argInfo = new ArgInfo(typeName, argText);
            argInfos.push(argInfo);
        }
        var defaultHelpGroup = "";
        if (typeof keyword === "string") {
            keyword = [[keyword], defaultHelpGroup];
        }
        else if (isNameList(keyword)) {
            keyword = [keyword, defaultHelpGroup];
        }
        var names = keyword[0];
        var helpGroup = keyword[1];
        var option = new Option(names, argInfos, text, helpGroup, func);
        var arity = argInfos.length;
        console.assert(func.length <= arity);
        var options = this._optionsByArity.get(arity);
        if (options === undefined) {
            options = [];
            this._optionsByArity.set(arity, options);
            this._maxArity = Math.max(this._maxArity, arity);
        }
        // TODO: Check for conflicting keywords.
        options.push(option);
    };
    OptionsImpl.prototype.forEachOption = function (func) {
        for (var arity = 0; arity <= this._maxArity; ++arity) {
            var options = this._optionsByArity.get(arity);
            if (options !== undefined) {
                for (var _i = 0, options_1 = options; _i < options_1.length; _i++) {
                    var option = options_1[_i];
                    if (func(option)) {
                        return true;
                    }
                }
            }
        }
        return false;
    };
    OptionsImpl.prototype.getHelpDescription = function (config) {
        var _this = this;
        var options = [];
        for (var arity = 0; arity <= this._maxArity; ++arity) {
            var opts = this._optionsByArity.get(arity);
            if (opts !== undefined) {
                for (var _i = 0, opts_1 = opts; _i < opts_1.length; _i++) {
                    var o = opts_1[_i];
                    options.push(o);
                }
            }
        }
        var getPriority = function (helpGroup) {
            var priority = _this._helpGroupPriorities.get(helpGroup);
            if (priority === undefined) {
                return 0;
            }
            return priority;
        };
        stable_1.inplace(options, function (opt1, opt2) {
            var g1 = opt1.helpGroup;
            var g2 = opt2.helpGroup;
            var p1 = getPriority(g1);
            var p2 = getPriority(g2);
            if (p1 < p2) {
                return -1;
            }
            if (p1 > p2) {
                return 1;
            }
            if (opt1.names.length === 0) {
                if (opt2.names.length === 0) {
                    return 0;
                }
                return -1;
            }
            if (opt2.names.length === 0) {
                return 1;
            }
            var n1 = opt1.names[0];
            var n2 = opt2.names[0];
            if (n1 < n2) {
                return -1;
            }
            if (n1 > n2) {
                return 1;
            }
            return 0;
        });
        var formatter = new Formatter(config);
        for (var _a = 0, options_2 = options; _a < options_2.length; _a++) {
            var option = options_2[_a];
            formatter.formatOption(option);
        }
        return formatter.toString();
    };
    OptionsImpl.prototype.setHelpGroupPriority = function (helpGroup, priority) {
        this._helpGroupPriorities.set(helpGroup, priority);
    };
    return OptionsImpl;
}());
var Exception = /** @class */ (function (_super) {
    __extends(Exception, _super);
    function Exception(message) {
        var _this = _super.call(this, message) || this;
        Object.setPrototypeOf(_this, Exception.prototype);
        return _this;
    }
    return Exception;
}(Error));
exports.Exception = Exception;
var RegistrationException = /** @class */ (function (_super) {
    __extends(RegistrationException, _super);
    function RegistrationException(conflictingTypeName) {
        var _this = _super.call(this, "Cannot register `" + conflictingTypeName + "': already exists.") || this;
        Object.setPrototypeOf(_this, RegistrationException.prototype);
        _this.conflictingTypeName = conflictingTypeName;
        return _this;
    }
    return RegistrationException;
}(Exception));
exports.RegistrationException = RegistrationException;
var OptionsStateException = /** @class */ (function (_super) {
    __extends(OptionsStateException, _super);
    function OptionsStateException(message) {
        var _this = _super.call(this, message) || this;
        Object.setPrototypeOf(_this, OptionsStateException.prototype);
        return _this;
    }
    return OptionsStateException;
}(Exception));
exports.OptionsStateException = OptionsStateException;
var ParseException = /** @class */ (function (_super) {
    __extends(ParseException, _super);
    function ParseException(args, message) {
        var _this = _super.call(this, message) || this;
        Object.setPrototypeOf(_this, ParseException.prototype);
        _this.args = args.slice();
        return _this;
    }
    ParseException.create = function (beginArgsIndex, endArgsIndex, args) {
        if (endArgsIndex === beginArgsIndex) {
            return new UnknownOptionException(beginArgsIndex, args);
        }
        else if (endArgsIndex === args.length) {
            return new EndOfInputException(beginArgsIndex, args);
        }
        else {
            return new BadInputException(beginArgsIndex, endArgsIndex, args);
        }
    };
    return ParseException;
}(Exception));
exports.ParseException = ParseException;
var BadInputException = /** @class */ (function (_super) {
    __extends(BadInputException, _super);
    function BadInputException(keywordArgsIndex, badArgsIndex, args) {
        var _this = _super.call(this, args, "Bad input for `" + args[keywordArgsIndex] + "' at index " + badArgsIndex + ": `" + args[badArgsIndex] + "'.") || this;
        Object.setPrototypeOf(_this, BadInputException.prototype);
        _this.keywordArgsIndex = keywordArgsIndex;
        _this.badArgsIndex = badArgsIndex;
        return _this;
    }
    return BadInputException;
}(ParseException));
exports.BadInputException = BadInputException;
var UnknownOptionException = /** @class */ (function (_super) {
    __extends(UnknownOptionException, _super);
    function UnknownOptionException(keywordArgsIndex, args) {
        var _this = _super.call(this, args, "Unknown option at index " + keywordArgsIndex + ": `" + args[keywordArgsIndex] + "'.") || this;
        Object.setPrototypeOf(_this, UnknownOptionException.prototype);
        _this.keywordArgsIndex = keywordArgsIndex;
        return _this;
    }
    return UnknownOptionException;
}(ParseException));
exports.UnknownOptionException = UnknownOptionException;
var EndOfInputException = /** @class */ (function (_super) {
    __extends(EndOfInputException, _super);
    function EndOfInputException(keywordArgsIndex, args) {
        var _this = _super.call(this, args, "Bad input for `" + args[keywordArgsIndex] + "' at index " + args[args.length] + ": End of input.") || this;
        Object.setPrototypeOf(_this, EndOfInputException.prototype);
        _this.keywordArgsIndex = keywordArgsIndex;
        return _this;
    }
    return EndOfInputException;
}(ParseException));
exports.EndOfInputException = EndOfInputException;
var ParseContext = /** @class */ (function () {
    function ParseContext(impl, args) {
        this._impl = impl;
        this._args = args.slice();
    }
    ParseContext.prototype.run = function () {
        var error = this._tryParse();
        if (error === null) {
            return;
        }
        throw error;
    };
    ParseContext.prototype._matchOptionKeyword = function (option, argsIndex) {
        if (option.names.length === 0) {
            return 0;
        }
        var arg = this._args[argsIndex];
        for (var _i = 0, _a = option.names; _i < _a.length; _i++) {
            var name_1 = _a[_i];
            if (name_1 === arg) {
                return 1;
            }
        }
        return null;
    };
    ParseContext.prototype._tryParse = function () {
        var _this = this;
        var argsIndex = 0;
        var highArgsIndex = argsIndex;
        var parsedItemBuffer = [];
        var lazyCalls = [];
        while (argsIndex < this._args.length) {
            var parsedAnOption = this._impl.forEachOption(function (option) {
                var keywordConsumedCount = _this._matchOptionKeyword(option, argsIndex);
                if (keywordConsumedCount === null) {
                    return false;
                }
                var localArgsIndex = argsIndex + keywordConsumedCount;
                highArgsIndex = Math.max(highArgsIndex, localArgsIndex);
                for (var i = 0; i < option.argInfos.length; ++i) {
                    var argInfo = option.argInfos[i];
                    var parser = _this._impl.getParser(argInfo.typeName);
                    var _a = parser(localArgsIndex, _this._args), parsedItem = _a[0], consumedCount = _a[1];
                    highArgsIndex = Math.max(highArgsIndex, localArgsIndex + consumedCount);
                    if (parsedItem === exports.NoParse) {
                        return false;
                    }
                    else {
                        localArgsIndex += consumedCount;
                        parsedItemBuffer[i] = parsedItem;
                    }
                }
                var parsedItems = parsedItemBuffer.slice(0, localArgsIndex - argsIndex);
                lazyCalls.push(function () {
                    option.func.apply(null, parsedItems);
                });
                argsIndex = localArgsIndex;
                return true;
            });
            if (!parsedAnOption) {
                return ParseException.create(argsIndex, highArgsIndex, this._args);
            }
        }
        console.assert(argsIndex === this._args.length);
        for (var _i = 0, lazyCalls_1 = lazyCalls; _i < lazyCalls_1.length; _i++) {
            var lazyCall = lazyCalls_1[_i];
            lazyCall();
        }
        return null;
    };
    return ParseContext;
}());
exports.ParseContext = ParseContext;
var Options = /** @class */ (function () {
    function Options(impl) {
        this._spawnedParseContexts = false;
        this._impl = impl;
    }
    Options.createRaw = function () {
        var impl = OptionsImpl.createRaw();
        return new Options(impl);
    };
    Options.createDefault = function () {
        return this.createRaw()
            .registerParsers("boolean", "boolean?", "boolean[]", parseBoolean)
            .registerParsers("string", "string?", "string[]", parseString)
            .registerParsers("number", "number?", "number[]", parseNumber)
            .registerParsers("integer", "integer?", "integer[]", parseInteger)
            .registerParsers("unsigned", "unsigned?", "unsigned[]", parseUnsigned)
            .registerParsers("Int8", "Int8?", "Int8[]", exports.parseInt8)
            .registerParsers("Int16", "Int16?", "Int16[]", exports.parseInt16)
            .registerParsers("Int32", "Int32?", "Int32[]", exports.parseInt32)
            .registerParsers("Uint8", "Uint8?", "Uint8[]", exports.parseUint8)
            .registerParsers("Uint16", "Uint16?", "Uint16[]", exports.parseUint16)
            .registerParsers("Uint32", "Uint32?", "Uint32[]", exports.parseUint32)
            .registerParser("Int8Array", exports.parseInt8Array)
            .registerParser("Int16Array", exports.parseInt16Array)
            .registerParser("Int32Array", exports.parseInt32Array)
            .registerParser("Uint8Array", exports.parseUint8Array)
            .registerParser("Uint16Array", exports.parseUint16Array)
            .registerParser("Uint32Array", exports.parseUint32Array);
    };
    Options.prototype.registerParsers = function (typeName, optionalTypeName, arrayTypeName, simpleParser) {
        if (optionalTypeName !== typeName + "?") {
            throw new TypeError("Expected `optionalTypeName' to be \"" + typeName + "?\" instead of \"" + optionalTypeName + "\"");
        }
        if (arrayTypeName !== typeName + "[]") {
            throw new TypeError("Expected `optionalTypeName' to be \"" + typeName + "[]\" instead of \"" + optionalTypeName + "\"");
        }
        var parser = simpleParse(simpleParser);
        var optionalParser = createOptionalParser(parser);
        var arrayParser = createArrayParser(parser);
        return this
            .registerParser(typeName, parser)
            .registerParser(optionalTypeName, optionalParser)
            .registerParser(arrayTypeName, arrayParser);
    };
    Options._cannotCallException = function (caller, invalidatingCall) {
        return new OptionsStateException("Cannot call `" + caller + "' after `" + invalidatingCall + "' is called.");
    };
    Options._nullImplException = function (caller) {
        return this._cannotCallException(caller, "registerParser");
    };
    // XXX:
    // Is it possible to restrict [[TypeName]] to exactly one string type?
    // As it is, the type system does not prevent one from doing
    // [[ options.registerParser(String(Math.random()), simpleParse(id)); ]]
    // The problem with this is that [[ TypeName ==== string ]]
    // Even doing the following is not desirable:
    // [[ options.registerParser(s as ("hello" | "world"), simpleParse(id)); ]]
    // Fortunately this shouldn't be a problem in practice, as people are
    // usually going to type in string literals, as it is natural to do so.
    Options.prototype.registerParser = function (typeName, parser) {
        if (this._impl === null) {
            throw Options._nullImplException("registerParser");
        }
        var oldImpl = this._impl;
        this._impl = null;
        var newImpl = oldImpl.registerParser(typeName, parser);
        return new Options(newImpl);
    };
    Options.prototype.setHelpGroupPriority = function (helpGroup, priority) {
        if (this._impl === null) {
            throw Options._nullImplException("setHelpGroupPriority");
        }
        this._impl.setHelpGroupPriority(helpGroup, priority);
    };
    Options.prototype.addOption = function (keyword) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        if (this._impl === null) {
            throw Options._nullImplException("addOption");
        }
        if (this._spawnedParseContexts) {
            throw Options._cannotCallException("addOption", "createParseContext");
        }
        (_a = this._impl).addOption.apply(_a, [keyword].concat(args));
        var _a;
    };
    Options.queryStringToArgs = function (queryString, arrayFormat) {
        if (arrayFormat === void 0) { arrayFormat = "none"; }
        var config = { arrayFormat: arrayFormat };
        var dict = query_string_1.default.parse(queryString, config);
        var keys = Object.keys(dict);
        var args = [];
        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
            var key = keys_1[_i];
            var value = dict[key];
            if (typeof value === "string") {
                args.push(key, value);
            }
            else {
                for (var _a = 0, value_1 = value; _a < value_1.length; _a++) {
                    var s = value_1[_a];
                    args.push(key, s);
                }
            }
        }
        return args;
    };
    Options.prototype.createParseContext = function (args) {
        if (this._impl === null) {
            throw Options._nullImplException("createParseContext");
        }
        this._spawnedParseContexts = true;
        return new ParseContext(this._impl, args);
    };
    Options.prototype.getHelpDescription = function (config) {
        if (this._impl === null) {
            throw Options._nullImplException("getHelpDescription");
        }
        if (config === undefined) {
            config = {
                maxWidth: 80,
                helpGroupFilter: [],
            };
        }
        return this._impl.getHelpDescription(config);
    };
    return Options;
}());
exports.Options = Options;
